#define PROGRAM_FILE "initmt.cl"
#define PROGRAM_FILE2 "update.cl"
#define PROGRAM_FILE3 "swap.cl"
#define KERNEL_FUNC "initmt"
#define KERNEL_FUNC2 "update"
#define KERNEL_FUNC3 "swap"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#ifdef MAC
#include <OpenCL/cl.h>
#else
#include <CL/cl.h>
#endif
/* Find a device associated with the first available platform */
cl_device_id create_device() {

   cl_platform_id platform;
   cl_device_id dev[10],device;
   cl_uint number,i;
   int err;

   /* Identify a platform */
   err = clGetPlatformIDs(1, &platform, &number);
   if(err < 0) {
      perror("Couldn't identify a platform");
      exit(1);
   } 
   char buffer[10240]; 
   cl_uint buf_uint,idims;
   cl_ulong buf_ulong;
   size_t wgsize,wisize[3];
   
   for(i=0; i<number; i++) {
       clGetPlatformInfo(platform,CL_PLATFORM_PROFILE,10240,buffer,NULL);
       printf("PROFILE=%s\n",buffer);
       clGetPlatformInfo(platform,CL_PLATFORM_VERSION,10240,buffer,NULL);
       printf("VERSION=%s\n",buffer);
       clGetPlatformInfo(platform,CL_PLATFORM_NAME,10240,buffer,NULL);
       printf("NAME=%s\n",buffer);
       clGetPlatformInfo(platform,CL_PLATFORM_VENDOR,10240,buffer,NULL);
       printf("VENDOR=%s\n",buffer);
       clGetPlatformInfo(platform,CL_PLATFORM_EXTENSIONS,10240,buffer,NULL);
       printf("EXTENSIONS=%s\n",buffer);
   }

   /* Access a device */
   err = clGetDeviceIDs(platform, CL_DEVICE_TYPE_ALL, 10, dev, &number);
   if(err == CL_DEVICE_NOT_FOUND) {
      perror("Couldn't access any devices");
      exit(1);   
   }
   for(i=0; i<number; i++) {
       printf("--%d--\n",i); 
       clGetDeviceInfo(dev[i],CL_DEVICE_NAME,10240,buffer,NULL);
       printf("DEVICE NAME=%s\n",buffer);
       clGetDeviceInfo(dev[i],CL_DEVICE_VENDOR,10240,buffer,NULL);
       printf("DEVICE VENDOR=%s\n",buffer);
       clGetDeviceInfo(dev[i],CL_DEVICE_VERSION,10240,buffer,NULL);
       printf("DEVICE VERSION=%s\n",buffer);
       clGetDeviceInfo(dev[i],CL_DEVICE_MAX_COMPUTE_UNITS,sizeof(buf_uint),&buf_uint,NULL);
       printf("DEVICE_MAX_COMPUTE_UNITS=%u\n",buf_uint);
       clGetDeviceInfo(dev[i],CL_DEVICE_MAX_WORK_GROUP_SIZE,sizeof(size_t),&wgsize,NULL);
       printf("DEVICE_MAX_WORK_GROUP_SIZE=%i\n",(int)wgsize);
       clGetDeviceInfo(dev[i],CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS,sizeof(cl_uint),&idims,NULL);
       printf("DEVICE_MAX_WORK_ITEM_DIMENSIONS=%u\n",idims);
       clGetDeviceInfo(dev[i],CL_DEVICE_MAX_WORK_ITEM_SIZES,sizeof(size_t[3]),&wisize,NULL);
       printf("DEVICE_MAX_WORK_ITEM_SIZES=%lu %lu %lu\n",wisize[0],wisize[1],wisize[2]);
       clGetDeviceInfo(dev[i],CL_DEVICE_GLOBAL_MEM_SIZE,sizeof(cl_ulong),&buf_ulong,NULL);
       printf("DEVICE_GLOBAL_MEM_SIZE=%lu\n",buf_ulong);
   }
   device=dev[1]; 
   return device;
}

/* Create program from a file and compile it */
cl_program build_program(cl_context ctx, cl_device_id dev, const char* filename) {

   cl_program program;
   FILE *program_handle;
   char *program_buffer, *program_log;
   size_t program_size, log_size;
   int err;
   char options[1000];


   /* Read program file and place content into buffer */
   program_handle = fopen(filename, "r");
   if(program_handle == NULL) {
      perror("Couldn't find the program file");
      exit(1);
   }
   fseek(program_handle, 0, SEEK_END);
   program_size = ftell(program_handle);
   rewind(program_handle);
   program_buffer = (char*)malloc(program_size + 1);
   program_buffer[program_size] = '\0';
   fread(program_buffer, sizeof(char), program_size, program_handle);
   fclose(program_handle);

   /* Create program from file */
   program = clCreateProgramWithSource(ctx, 1, 
      (const char**)&program_buffer, &program_size, &err);
   if(err < 0) {
      perror("Couldn't create the program");
      exit(1);
   }
   free(program_buffer);
   strcpy(options, "");
   strcat(options, "-cl-mad-enable ");
#ifdef INTEL_MIC
   strcat(options, "-auto-prefetch-level=3 ");
#endif
   fprintf(stdout, "Build OpenCL (opts=\"%s\")\n", options);
 
   /* Build program */
   err = clBuildProgram(program, 0, NULL, options, NULL, NULL);
   if(err < 0) {

      /* Find size of log and print to std output */
      clGetProgramBuildInfo(program, dev, CL_PROGRAM_BUILD_LOG, 
            0, NULL, &log_size);
      program_log = (char*) malloc(log_size + 1);
      program_log[log_size] = '\0';
      clGetProgramBuildInfo(program, dev, CL_PROGRAM_BUILD_LOG, 
            log_size + 1, program_log, NULL);
      printf("%s\n", program_log);
      free(program_log);
      exit(1);
   }

   return program;
}

/********************************************************************

 This benchmark test program is measuring a cpu performance
 of floating point operation by a Poisson equation solver.

 If you have any question, please ask me via email.
 written by Ryutaro HIMENO, November 26, 2001.
 Version 3.0
 ----------------------------------------------
 Ryutaro Himeno, Dr. of Eng.
 Head of Computer Information Division,
 RIKEN (The Institute of Pysical and Chemical Research)
 Email : himeno@postman.riken.go.jp
 ---------------------------------------------------------------
 You can adjust the size of this benchmark code to fit your target
 computer. In that case, please chose following sets of
 (mimax,mjmax,mkmax):
 small : 33,33,65
 small : 65,65,129
 midium: 129,129,257
 large : 257,257,513
 ext.large: 513,513,1025
 This program is to measure a computer performance in MFLOPS
 by using a kernel which appears in a linear solver of pressure
 Poisson eq. which appears in an incompressible Navier-Stokes solver.
 A point-Jacobi method is employed in this solver as this method can 
 be easyly vectrized and be parallelized.
 ------------------
 Finite-difference method, curvilinear coodinate system
 Vectorizable and parallelizable on each grid point
 No. of grid points : imax x jmax x kmax including boundaries
 ------------------
 A,B,C:coefficient matrix, wrk1: source term of Poisson equation
 wrk2 : working area, OMEGA : relaxation parameter
 BND:control variable for boundaries and objects ( = 0 or 1)
 P: pressure
********************************************************************/

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef SSMALL
#define MIMAX            33
#define MJMAX            33
#define MKMAX            65
#endif

#ifdef SMALL
#define MIMAX            65
#define MJMAX            65
#define MKMAX            129
#endif

#ifdef MIDDLE
#define MIMAX            129
#define MJMAX            129
#define MKMAX            257
#define KERNELMIMAX      144
#define KERNELMJMAX      144
#define KERNELMKMAX      288
#endif

#ifdef LARGE
#define MIMAX            257
#define MJMAX            257
#define MKMAX            513
#define KERNELMIMAX      288
#define KERNELMJMAX      288
#define KERNELMKMAX      544
#endif

#ifdef ELARGE
#define MIMAX            513
#define MJMAX            513
#define MKMAX            1025
#endif

double second();
float jacobi();
void  initmt(int, int, int,int,int,int,float*,float*,float*,float*,float*,float*,float*,float*,float*,float*,float*,float*,float*);
double fflop(int,int,int);
float summa(int,int,int,float*);
double mflops(int,double,double);

int imax, jmax, kmax;
static float omega;
float checkp;
int
main()
{

  /* OpenCL structures */
   cl_device_id device;
   cl_context context;
   cl_program program,program2,program3;
   cl_kernel kernel,kernel2,kernel3;
   cl_command_queue queue;
   cl_int i, j, err,counter;
   cl_mem a1_buffer,a2_buffer,a3_buffer,a4_buffer,b1_buffer,b2_buffer,b3_buffer,c1_buffer,c2_buffer,c3_buffer;
   cl_mem p_buffer,bnd_buffer,wrk1_buffer,wrk2_buffer;
   
#define idxzz(i,j,k) ((i)+MIMAX*((j)+MJMAX*(k)))
  int    k,nn,imezzi;
  float  gosa;
  double cpu,cpu0,cpu1,flop,target;
  float *p,*a1,*a2,*a3,*a4,*b1,*b2,*b3,*c1,*c2,*c3,*bnd,*wrk1,*wrk2;

  target= 60.0;
  omega= 0.8;

  imax = MIMAX-1;
  jmax = MJMAX-1;
  kmax = MKMAX-1;
  int mimax = MIMAX;
  int mjmax = MJMAX;
  int mkmax = MKMAX;

  /*
   *    Initializing matrixes
   */
  int msize=MIMAX*MJMAX*MKMAX;
  a1 = (float *)malloc(msize*sizeof(float));
  a2 = (float *)malloc(msize*sizeof(float));
  a3 = (float *)malloc(msize*sizeof(float));
  a4 = (float *)malloc(msize*sizeof(float));
  b1 = (float *)malloc(msize*sizeof(float));
  b2 = (float *)malloc(msize*sizeof(float));
  b3 = (float *)malloc(msize*sizeof(float));
  c1 = (float *)malloc(msize*sizeof(float));
  c2 = (float *)malloc(msize*sizeof(float));
  c3 = (float *)malloc(msize*sizeof(float));
  p = (float *)malloc(msize*sizeof(float));
  bnd = (float *)malloc(msize*sizeof(float));
  wrk1 = (float *)malloc(msize*sizeof(float));
  wrk2 = (float *)malloc(msize*sizeof(float));
  
  /* Create device and context */
   device = create_device();
   context = clCreateContext(NULL, 1, &device, NULL, NULL, &err);
   if(err < 0) {
      perror("Couldn't create a context");
      exit(1);
   }

   /* Build program */
   program = build_program(context, device, PROGRAM_FILE);


  /* Create data buffer */
   a1_buffer = clCreateBuffer(context, 
      CL_MEM_USE_HOST_PTR, msize * sizeof(float), a1, &err);
   if(err < 0) {
      perror("Couldn't create a1 buffer");
      exit(1);
   };
   a2_buffer = clCreateBuffer(context, 
         CL_MEM_USE_HOST_PTR, msize * sizeof(float), a2, &err);
   if(err < 0) {
      perror("Couldn't create a2 buffer");
      exit(1);
   };
   a3_buffer = clCreateBuffer(context, 
         CL_MEM_USE_HOST_PTR, msize * sizeof(float), a3, &err);
   if(err < 0) {
      perror("Couldn't create a3 buffer");
      exit(1);
   };
   a4_buffer = clCreateBuffer(context, 
         CL_MEM_USE_HOST_PTR, msize * sizeof(float), a4, &err);
   if(err < 0) {
      perror("Couldn't create a4 buffer");
      exit(1);
   };
   b1_buffer = clCreateBuffer(context, 
         CL_MEM_USE_HOST_PTR, msize * sizeof(float), b1, &err);
   if(err < 0) {
      perror("Couldn't create b1 buffer");
      exit(1);
   };
   b2_buffer = clCreateBuffer(context, 
         CL_MEM_USE_HOST_PTR, msize * sizeof(float), b2, &err);
   if(err < 0) {
      perror("Couldn't create b2 buffer");
      exit(1);
   };
   b3_buffer = clCreateBuffer(context, 
         CL_MEM_USE_HOST_PTR, msize * sizeof(float), b3, &err);
   if(err < 0) {
      perror("Couldn't create b3 buffer");
      exit(1);
   };
   c1_buffer = clCreateBuffer(context, 
         CL_MEM_USE_HOST_PTR, msize * sizeof(float), c1, &err);
   if(err < 0) {
      perror("Couldn't create c1 buffer");
      exit(1);
   };
   c2_buffer = clCreateBuffer(context, 
         CL_MEM_USE_HOST_PTR, msize * sizeof(float), c2, &err);
   if(err < 0) {
      perror("Couldn't create c2 buffer");
      exit(1);
   };
   c3_buffer = clCreateBuffer(context, 
         CL_MEM_USE_HOST_PTR, msize * sizeof(float), c3, &err);
   if(err < 0) {
      perror("Couldn't create c3 buffer");
      exit(1);
   };
   p_buffer = clCreateBuffer(context, 
         CL_MEM_USE_HOST_PTR, msize * sizeof(float), p, &err);
   if(err < 0) {
      perror("Couldn't create p buffer");
      exit(1);
   };
   bnd_buffer = clCreateBuffer(context, 
         CL_MEM_USE_HOST_PTR, msize * sizeof(float), bnd, &err);
   if(err < 0) {
      perror("Couldn't create bnd buffer");
      exit(1);
   };
   wrk1_buffer = clCreateBuffer(context, 
         CL_MEM_USE_HOST_PTR, msize * sizeof(float), wrk1, &err);
   if(err < 0) {
      perror("Couldn't create wrk1 buffer");
      exit(1);
   };
   wrk2_buffer = clCreateBuffer(context, 
         CL_MEM_USE_HOST_PTR, msize * sizeof(float), wrk2, &err);
   if(err < 0) {
      perror("Couldn't create wrk2 buffer");
      exit(1);
   };

  /* Create a command queue */
   queue = clCreateCommandQueue(context, device, 0, &err);
   if(err < 0) {
      perror("Couldn't create a command queue");
      exit(1);
   };

   /* Create a kernel */
   kernel = clCreateKernel(program, KERNEL_FUNC, &err);
   if(err < 0) {
      perror("Couldn't create a kernel");
      exit(1);
   };
   
   /* Create kernel arguments */
   err = clSetKernelArg(kernel, 0, sizeof(int), &mimax);
   err = clSetKernelArg(kernel, 1, sizeof(int), &mjmax);
   err = clSetKernelArg(kernel, 2, sizeof(int), &mkmax);
   err = clSetKernelArg(kernel, 3, sizeof(int), &imax);
   err = clSetKernelArg(kernel, 4, sizeof(int), &jmax);
   err = clSetKernelArg(kernel, 5, sizeof(int), &kmax);
   err = clSetKernelArg(kernel, 6, sizeof(cl_mem), &a1_buffer);
   err = clSetKernelArg(kernel, 7, sizeof(cl_mem), &a2_buffer);
   err = clSetKernelArg(kernel, 8, sizeof(cl_mem), &a3_buffer);
   err = clSetKernelArg(kernel, 9, sizeof(cl_mem), &a4_buffer);
   err = clSetKernelArg(kernel, 10, sizeof(cl_mem), &b1_buffer);
   err = clSetKernelArg(kernel, 11, sizeof(cl_mem), &b2_buffer);
   err = clSetKernelArg(kernel, 12, sizeof(cl_mem), &b3_buffer);
   err = clSetKernelArg(kernel, 13, sizeof(cl_mem), &c1_buffer);
   err = clSetKernelArg(kernel, 14, sizeof(cl_mem), &c2_buffer);
   err = clSetKernelArg(kernel, 15, sizeof(cl_mem), &c3_buffer);
   err = clSetKernelArg(kernel, 16, sizeof(cl_mem), &p_buffer);
   err = clSetKernelArg(kernel, 17, sizeof(cl_mem), &wrk1_buffer);
   err = clSetKernelArg(kernel, 18, sizeof(cl_mem), &bnd_buffer);
   //if(err < 0) { perror("Couldn't create a kernel argument"); exit(1); }
  const size_t gws[3]={KERNELMIMAX,KERNELMJMAX,KERNELMKMAX};
  const size_t lws[3]={1,16,32};
  printf("global work size = %d %d %d\n",gws[0],gws[1],gws[2]);
  printf("local work size = %d %d %d\n",lws[0],lws[1],lws[2]);

  /* Enqueue kernel */
   err = clEnqueueNDRangeKernel(queue, kernel, 3, NULL, gws,
         NULL, 0, NULL, NULL);
   if(err < 0) {
      perror("Couldn't enqueue the kernel 1");
  printf("Error %d\n",err);
      exit(1);
   }

  //initmt(MIMAX,MJMAX,MKMAX,imax,jmax,kmax,a1,a2,a3,a4,b1,b2,b3,c1,c2,c3,p,wrk1,bnd);
  printf("mimax = %d mjmax = %d mkmax = %d\n",MIMAX, MJMAX, MKMAX);
  printf("imax = %d jmax = %d kmax =%d\n",imax,jmax,kmax);
  nn= 100;
  printf(" Start Himeno measurement process.\n");
  printf(" Measure the performance in %d times.\n\n",nn);
   /* Build program */
   program2 = build_program(context, device, PROGRAM_FILE2);
   /* Create a kernel */
   kernel2 = clCreateKernel(program2, KERNEL_FUNC2, &err);
   if(err < 0) {
      perror("Couldn't create a kernel");
      exit(1);
   };
   /* Create kernel arguments */
   err = clSetKernelArg(kernel2, 0, sizeof(float), &omega);
   err = clSetKernelArg(kernel2, 1, sizeof(int), &mimax);
   err = clSetKernelArg(kernel2, 2, sizeof(int), &mjmax);
   err = clSetKernelArg(kernel2, 3, sizeof(int), &mkmax);
   err = clSetKernelArg(kernel2, 4, sizeof(int), &imax);
   err = clSetKernelArg(kernel2, 5, sizeof(int), &jmax);
   err = clSetKernelArg(kernel2, 6, sizeof(int), &kmax);
   err = clSetKernelArg(kernel2, 7, sizeof(cl_mem), &a1_buffer);
   err = clSetKernelArg(kernel2, 8, sizeof(cl_mem), &a2_buffer);
   err = clSetKernelArg(kernel2, 9, sizeof(cl_mem), &a3_buffer);
   err = clSetKernelArg(kernel2, 10, sizeof(cl_mem), &a4_buffer);
   err = clSetKernelArg(kernel2, 11, sizeof(cl_mem), &b1_buffer);
   err = clSetKernelArg(kernel2, 12, sizeof(cl_mem), &b2_buffer);
   err = clSetKernelArg(kernel2, 13, sizeof(cl_mem), &b3_buffer);
   err = clSetKernelArg(kernel2, 14, sizeof(cl_mem), &c1_buffer);
   err = clSetKernelArg(kernel2, 15, sizeof(cl_mem), &c2_buffer);
   err = clSetKernelArg(kernel2, 16, sizeof(cl_mem), &c3_buffer);
   err = clSetKernelArg(kernel2, 17, sizeof(cl_mem), &wrk1_buffer);
   err = clSetKernelArg(kernel2, 18, sizeof(cl_mem), &wrk2_buffer);
   err = clSetKernelArg(kernel2, 19, sizeof(cl_mem), &bnd_buffer);
   err = clSetKernelArg(kernel2, 20, sizeof(cl_mem), &p_buffer);
  /* Build program */
   program3 = build_program(context, device, PROGRAM_FILE3);
   /* Create a kernel */
   kernel3 = clCreateKernel(program3, KERNEL_FUNC3, &err);
   if(err < 0) {
      perror("Couldn't create a kernel");
      exit(1);
   };
   /* Create kernel arguments */
   err = clSetKernelArg(kernel3, 0, sizeof(int), &mimax);
   err = clSetKernelArg(kernel3, 1, sizeof(int), &mjmax);
   err = clSetKernelArg(kernel3, 2, sizeof(int), &mkmax);
   err = clSetKernelArg(kernel3, 3, sizeof(int), &imax);
   err = clSetKernelArg(kernel3, 4, sizeof(int), &jmax);
   err = clSetKernelArg(kernel3, 5, sizeof(int), &kmax);
   err = clSetKernelArg(kernel3, 6, sizeof(cl_mem), &wrk2_buffer);
   err = clSetKernelArg(kernel3, 7, sizeof(cl_mem), &p_buffer);


  cpu0= second();
   for(i=0; i<nn; i++) {
  /* Enqueue kernel */
   err = clEnqueueNDRangeKernel(queue, kernel2, 3, NULL, gws,
         NULL, 0, NULL, NULL);
   if(err < 0) {
      perror("Couldn't enqueue the kernel");
      exit(1);
   }
  /* Enqueue kernel */
   err = clEnqueueNDRangeKernel(queue, kernel3, 3, NULL, gws,
         NULL, 0, NULL, NULL);
   if(err < 0) {
      perror("Couldn't enqueue the kernel");
      exit(1);
   }
   }
   /* Read the kernel's output */
   err = clEnqueueReadBuffer(queue, p_buffer, CL_TRUE, 0,
         msize*sizeof(float), p, 0, NULL, NULL);
   if(err < 0) {
      perror("Couldn't read p buffer");
      exit(1);
   }
     /* End of read the kernel's output */
  cpu1=second();cpu=cpu1-cpu0;
  flop= fflop(imax,jmax,kmax);
  printf(" MFLOPS: %f time(s): %f %e\n\n",
  mflops(nn,cpu,flop),cpu,gosa);
  checkp=summa(MIMAX, MJMAX,MKMAX,p );
  printf(" sum p: %f\n",checkp);
  imezzi=idxzz(imax/2,jmax/2,kmax/2);
  printf(" p mezzi: %f\n",p[imezzi]);

  return (0);
}

void
initmt(int MIMX,int MJMX,int MKMX,int imax,int jmax,int kmax,float* a1,float* a2,float* a3,float* a4,float* b1,float* b2,float* b3,float* c1,float* c2,float* c3,float* p,float* wrk1,float* bnd)
{
#define idxz(i,j,k) ((k)+MKMX*((j)+MJMX*(i)))
	int i,j,k;
  for(i=0 ; i<MIMX ; i++)
    for(j=0 ; j<MJMX ; j++)
      for(k=0 ; k<MKMX ; k++){
        a1[idxz(i,j,k)]=0.0;
        a2[idxz(i,j,k)]=0.0;
        a3[idxz(i,j,k)]=0.0;
        a4[idxz(i,j,k)]=0.0;
        b1[idxz(i,j,k)]=0.0;
        b2[idxz(i,j,k)]=0.0;
        b3[idxz(i,j,k)]=0.0;
        c1[idxz(i,j,k)]=0.0;
        c2[idxz(i,j,k)]=0.0;
        c3[idxz(i,j,k)]=0.0;
        p[idxz(i,j,k)]=0.0;
        wrk1[idxz(i,j,k)]=0.0;
        bnd[idxz(i,j,k)]=0.0;
      }

  for(i=0 ; i<imax ; i++)
    for(j=0 ; j<jmax ; j++)
      for(k=0 ; k<kmax ; k++){
        a1[idxz(i,j,k)]=1.0;
        a2[idxz(i,j,k)]=1.0;
        a3[idxz(i,j,k)]=1.0;
        a4[idxz(i,j,k)]=1.0/6.0;
        b1[idxz(i,j,k)]=0.0;
        b2[idxz(i,j,k)]=0.0;
        b3[idxz(i,j,k)]=0.0;
        c1[idxz(i,j,k)]=1.0;
        c2[idxz(i,j,k)]=1.0;
        c3[idxz(i,j,k)]=1.0;
        p[idxz(i,j,k)]=(float)(i*i)/(float)((imax-1)*(imax-1));
        wrk1[idxz(i,j,k)]=0.0;
        bnd[idxz(i,j,k)]=1.0;
      }
}

float
jacobi(int nn,int MIMX,int MJMX,int MKMX,int imax,int jmax,int kmax,float* a1,float* a2,float* a3,float* a4,float* b1,float* b2,float* b3,float* c1,float* c2,float* c3,float* p,float* wrk1,float* bnd,float* wrk2)
{
#define idxz(i,j,k) ((k)+MKMX*((j)+MJMX*(i)))
  int i,j,k,n;
  float gosa, s0, ss;

  for(n=0 ; n<nn ; ++n){
    gosa = 0.0;

    for(i=1 ; i<imax-1 ; i++)
      for(j=1 ; j<jmax-1 ; j++)
        for(k=1 ; k<kmax-1 ; k++){
          s0 = a1[idxz(i,j,k)] * p[idxz(i+1,j,k)]
             + a2[idxz(i,j,k)] * p[idxz(i,j+1,k)]
             + a3[idxz(i,j,k)] * p[idxz(i,j,k+1)]
             + b1[idxz(i,j,k)] * ( p[idxz(i+1,j+1,k)] - p[idxz(i+1,j-1,k)]
                              - p[idxz(i-1,j+1,k)  ] + p[idxz(i-1,j-1,k)  ] )
             + b2[idxz(i,j,k)] * ( p[idxz(i,j+1,k+1)] - p[idxz(i,j-1,k+1)]
                               - p[idxz(i,j+1,k-1)] + p[idxz(i,j-1,k-1)] )
             + b3[idxz(i,j,k)] * ( p[idxz(i+1,j,k+1)] - p[idxz(i-1,j,k+1)]
                               - p[idxz(i+1,j,k-1)] + p[idxz(i-1,j,k-1)] )
             + c1[idxz(i,j,k)] * p[idxz(i-1,j,k)  ]
             + c2[idxz(i,j,k)] * p[idxz(i,j-1,k)  ]
             + c3[idxz(i,j,k)] * p[idxz(i,j,k-1)]
             + wrk1[idxz(i,j,k)];

          ss = ( s0 * a4[idxz(i,j,k)] - p[idxz(i,j,k)] ) * bnd[idxz(i,j,k)];

          gosa+= ss*ss;
          /* gosa= (gosa > ss*ss) ? a : b; */

          wrk2[idxz(i,j,k)] = p[idxz(i,j,k)] + omega * ss;
        }

    for(i=1 ; i<imax-1 ; ++i)
      for(j=1 ; j<jmax-1 ; ++j)
        for(k=1 ; k<kmax-1 ; ++k)
          p[idxz(i,j,k)] = wrk2[idxz(i,j,k)];
  } /* end n loop */

  return(0);
}

double
fflop(int mx,int my, int mz)
{
  return((double)(mz-2)*(double)(my-2)*(double)(mx-2)*34.0);
}

float
summa(int MI, int MJ, int MK,float* val )
{
float summa=0.0;
int i,j,k; 
  for(i=0 ; i<MI ; i++)
    for(j=0 ; j<MJ ; j++)
      for(k=0 ; k<MK ; k++){
#define idxzz(i,j,k) ((i)+MI*((j)+MJ*(k)))
          summa+=val[idxzz(i,j,k)];
 } /* end n loop */

  return(summa);
}

double
mflops(int nn,double cpu,double flop)
{
  return(flop/cpu*1.e-6*(double)nn);
}

double
second()
{
#include <sys/time.h>

  struct timeval tm;
  double t ;

  static int base_sec = 0,base_usec = 0;

  gettimeofday(&tm, NULL);
  
  if(base_sec == 0 && base_usec == 0)
    {
      base_sec = tm.tv_sec;
      base_usec = tm.tv_usec;
      t = 0.0;
  } else {
    t = (double) (tm.tv_sec-base_sec) + 
      ((double) (tm.tv_usec-base_usec))/1.0e6 ;
  }

  return t ;
}
