#include <stdlib.h>
#include <sys/types.h>

#ifdef WIN32
#include <winsock2.h>
#else
#include <sys/time.h>
#endif

/*  -------------------------------------------------------------------

    This function returns the wall clock time with micro seconds
    accuracy.
    The data type of the returned value is "double".
    
    The function can be called from a FORTRAN module. The value
    returned by cclock_ and cclock should be of type REAL(Kind = 8).

    -------------------------------------------------------------------
*/

double cclock_( void )
{
#ifdef WIN32
	static int first = 1;
	static LARGE_INTEGER start;	
	LARGE_INTEGER stop;	
	static LARGE_INTEGER frequency;

	if (first) {
		int res = QueryPerformanceFrequency( &frequency );

		if (!res) {
			return -1.0e0;
		}
		else {
			first = 0;
		}

		QueryPerformanceCounter( &start );

		return 0.0e0;
	}

	QueryPerformanceCounter( &stop );	

	return (double)((double)(stop.QuadPart - start.QuadPart))/((double)frequency.QuadPart);

#else

   const  double  micro = 1.0e-06;    /* Conversion constant */
   static long    start = 0L, startu;
   struct timeval tp; 	              /* Structure used by gettimeofday */
   double         wall_time;          /* To hold the result */


   if ( gettimeofday( &tp, NULL) == -1 )
      wall_time = -1.0e0;
   else if( !start ) {
      start  = tp.tv_sec;
      startu = tp.tv_usec;
      wall_time = 0.0e0;
   }
   else
      wall_time = (double) (tp.tv_sec - start) + micro*(tp.tv_usec - startu);

   return wall_time;
#endif
}


double cclock( void )
{
   return cclock_();
}
