#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <float.h>
#include <math.h>

#include "cfft4_opencl.h"

//#define TEST
#define FEPS DBL_EPSILON

int devType=CL_DEVICE_TYPE_ALL;
// OpenCL function for cfft4

void tic(struct timeval *from)
{
  gettimeofday(from,NULL);
}

double toc(struct timeval *from)
{
  struct timeval to;
  gettimeofday(&to,NULL);
  return (double)(1000000.0*(to.tv_sec-from->tv_sec) + to.tv_usec-from->tv_usec)/1000000.0;
}

//VICcl_device_id oclGetDev(cl_context cxGPUContext, unsigned int nr)
//VIC{
//VIC  size_t szParmDataBytes;
//VIC  cl_device_id* cdDevices;

  // get the list of GPU devices associated with context
//VIC  clGetContextInfo(cxGPUContext, CL_CONTEXT_DEVICES, 0, NULL, &szParmDataBytes);

//VIC  if( szParmDataBytes / sizeof(cl_device_id) < nr ) {
//VIC    return (cl_device_id)-1;
//VIC  }

//VIC  cdDevices = (cl_device_id*) malloc(szParmDataBytes);

//VIC  clGetContextInfo(cxGPUContext, CL_CONTEXT_DEVICES, szParmDataBytes, cdDevices, NULL);

//VIC  cl_device_id device = cdDevices[nr];
//VIC  free(cdDevices);

//VIC  return device;
//VIC}


void init_opencl(const int id)
{
  err = clGetPlatformIDs(1, &cpPlatform, NULL);
  if (err != CL_SUCCESS) {
    printf("Error: Failed to find a platform!\n");
    exit(1);
  }

  // Connect to a compute device
  err = clGetPlatformIDs(1, &cpPlatform, NULL);
  if (err != CL_SUCCESS) {
    printf("Error: Failed to find a platform!\n");
    exit(1);
  }

  // Get a device of the appropriate type
  cl_device_id dev[10],device;
  err = clGetDeviceIDs(cpPlatform, devType, 10, dev, &deviceCount);
  if (err != CL_SUCCESS) {
    printf("Error: Failed to create a device group!\n");
    exit(1);
  }

char buffer[10240];
cl_uint buf_uint,idims;
cl_ulong buf_ulong;
size_t wgsize,wisize[3];


for(int i=0; i<deviceCount; i++) {
       printf("--%d--\n",i);
       clGetDeviceInfo(dev[i],CL_DEVICE_NAME,10240,buffer,NULL);
       printf("DEVICE NAME=%s\n",buffer);
       clGetDeviceInfo(dev[i],CL_DEVICE_VENDOR,10240,buffer,NULL);
       printf("DEVICE VENDOR=%s\n",buffer);
       clGetDeviceInfo(dev[i],CL_DEVICE_VERSION,10240,buffer,NULL);
       printf("DEVICE VERSION=%s\n",buffer);
       clGetDeviceInfo(dev[i],CL_DEVICE_MAX_COMPUTE_UNITS,sizeof(buf_uint),&buf_uint,NULL);
       printf("DEVICE_MAX_COMPUTE_UNITS=%u\n",buf_uint);
       clGetDeviceInfo(dev[i],CL_DEVICE_MAX_WORK_GROUP_SIZE,sizeof(size_t),&wgsize,NULL);
       printf("DEVICE_MAX_WORK_GROUP_SIZE=%i\n",(int)wgsize);
       clGetDeviceInfo(dev[i],CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS,sizeof(cl_uint),&idims,NULL);
       printf("DEVICE_MAX_WORK_ITEM_DIMENSIONS=%u\n",idims);
       clGetDeviceInfo(dev[i],CL_DEVICE_MAX_WORK_ITEM_SIZES,sizeof(size_t[3]),&wisize,NULL);
       printf("DEVICE_MAX_WORK_ITEM_SIZES=%lu %lu %lu\n",wisize[0],wisize[1],wisize[2]);
       clGetDeviceInfo(dev[i],CL_DEVICE_GLOBAL_MEM_SIZE,sizeof(cl_ulong),&buf_ulong,NULL);

printf("DEVICE_GLOBAL_MEM_SIZE=%lu\n",buf_ulong);
   }


  // Create a compute context
  context = clCreateContext(0, deviceCount, dev, NULL, NULL, &err);
  if (!context) {
    printf("Error: Failed to create a compute context!\n");
    exit(1);
  }

  cdDevice = oclGetDev(context, id);

  // Create a command commands
  commands = clCreateCommandQueue(context, cdDevice, 0, &err);
  if (!commands) {
    printf("Error: Failed to create a command commands!\n");
    exit(1);
  }
  // get source code of kernel from file
  KernelSource = getKernelFromFile("cfft4_opencl_kernels.cl");
  // Create the compute program from the source buffer
  program = clCreateProgramWithSource(context, 1,
      (const char **) &KernelSource,
      NULL, &err);
  if (!program) {
    printf("Error: Failed to create compute program!\n");
    exit(1);
  }
  // Build the program executable
#ifdef MIC  
  err = clBuildProgram(program, 0, NULL, "-cl-fast-relaxed-math -auto-prefetch-level=3", NULL, NULL);
#else
  err = clBuildProgram(program, 0, NULL, "-cl-fast-relaxed-math", NULL, NULL);
#endif
   
  if (err != CL_SUCCESS) {
    size_t len;
    char buffer[2048];

    printf("Error: Failed to build program executable!\n");
    clGetProgramBuildInfo(program, cdDevice, CL_PROGRAM_BUILD_LOG,
        sizeof(buffer), buffer, &len);
    printf("%s\n", buffer);
    exit(1);
  }
}
void finish_opencl(void)
{
  clReleaseProgram(program);
  clReleaseCommandQueue(commands);
  clReleaseContext(context);
  free(KernelSource);
}

void cfft4_opencl(int   is,   int m,
                  double ur[], double ui[],
                  double xr[], double xi[],
                  double wr[], double wi[],
                  double *time_kernel, double *time_copy)
{
    /* ----------------------------------------------------------------------
       ! --- CFFT4 performs an N-point complex-to-complex FFT, where N = 2**M.
       !     On input XR holds the Real part of the complex array;
       !     XI holds the Imaginary part.
       !     On output XR contains the Real part of the transformed array.
       !     XI then contains the Imaginary part of the result.
       !     WR and WI are work arrays that store Real and Imaginary parts
       !     of intermediate results.
       !     UR and UI contain rotation factors for various stages of the
       !     transformation.
       !
       ! --- Before doing the actual transformation UR and UI must be
       !     initialised by calling CFFT4 with IS = 0, and M = MX, where MX
       !     is the maximum value of M that will be needed for any call of
       !     CFFT4.
       !
       ! --- A call to CFFT4 with IS = 1 (or -1) indicates a call to perform
       !     a FFT with positive (or negative) exponentials.
       !
       !     For any call to CFFT4 M must be M >= 2.
       !     Dimensions of the arrays XR, XI, UR, UI, WR, WI are:
       !
       !     XR[N], XI[N], WR[N], WI[N].
       !     UR[2**MX+1],       UI[2**MX+1]       for MX even.
       !     UR[2*2**(MX/3)+1], UI[2*2**(MX/3)+1] for MX odd.
       !
       !     In this routine we use 8-byte Real elements (doubles).
       ! ---------------------------------------------------------------------- */

  struct timeval tp;

  int no_blocks_r2 = 0;
  int no_blocks_r5 = 0;
  int n, k, mx, n1, l1, l2, l4, ln, ku, it;

#ifdef TEST
  int i, j, n2, n3 , l3, j0, j1, j2, j3, j4, j5 , k0,
      k1, k2, k3;

  double c0r, c0i, c1r, c1i, c2r, c2i, c3r, c3i,
        d0r, d0i, d1r, d1i, d2r, d2i, d3r, d3i;

  double u1r, u1i, u2r, u2i, u3r, u3i,
        w2, w1, x1, x2;

  int ok2=0;
#endif

  n = (int)pow( 2.0, m );

  cl_mem urcomp_gfxPtr;

  cl_mem uicomp_gfxPtr;

  // ---------------------------------------------------------------------
  // --- Initialise the U array with sines and cosines in a manner that
  //     permits stride one access at each FFT iteration.
  static double ur0=0.0;

  if ( is == 0 ) {
    int nu = (int)pow( 2.0, ((m+1)/2*2) )/3;

    ur0 = 64*nu+m;


  // Create the compute kernel in the program
  kernel = clCreateKernel(program, "cfft4_initU_kernel", &err);
  if (!kernel || err != CL_SUCCESS) {
    printf("Error: Failed to create compute kernel!\n");
    exit(1);
  }
  tic(&tp);
  // Create and copy to the device memory vectors
  urcomp_gfxPtr = clCreateBuffer(context, CL_MEM_READ_WRITE|CL_MEM_COPY_HOST_PTR ,
                                        sizeof(double) * n, ur, &err);
  if (err<0) {
    printf("Error: Failed to allocate device memory 1!\n");
    exit(1);
  }
  uicomp_gfxPtr = clCreateBuffer(context, CL_MEM_WRITE_ONLY|CL_MEM_COPY_HOST_PTR,
                                        sizeof(double) * n, ui, NULL);
  if (!uicomp_gfxPtr) {
    printf("Error: Failed to allocate device memory 11!\n");
    exit(1);
  }
  *time_copy+=toc(&tp);

  int nr_local = 128;
  int nr_threads = (2<<(2*( (m+1)/2 ))-1)/3 - 1;
  int no_blocks = nr_threads/nr_local;
  no_blocks += nr_threads%nr_local?1:0;

  if (no_blocks > 65535) {
    no_blocks = 65535;
  }

  tic(&tp);
  // Set the arguments to the compute kernel
  err = 0;
  err  = clSetKernelArg(kernel, 0, sizeof(int), &nr_threads);
  err |= clSetKernelArg(kernel, 1, sizeof(double), &ur0);
  err |= clSetKernelArg(kernel, 2, sizeof(cl_mem), (void*) &urcomp_gfxPtr);
  err |= clSetKernelArg(kernel, 3, sizeof(cl_mem), (void*) &uicomp_gfxPtr);
  if (err != CL_SUCCESS) {
    printf("Error: Failed to set kernel arguments!\n");
    exit(1);
  }

  size_t globalWorkers[] = {nr_local * no_blocks};
  size_t localWorkers[] = {nr_local};
  if (no_blocks != 0) {
    err = clEnqueueNDRangeKernel(commands ,kernel,1,0,globalWorkers,
        localWorkers,0,0, NULL);

    if (err != CL_SUCCESS) {
      printf("Error: Failed to execute kernel! %s\n", print_cl_errstring(err));
      exit(1);
    }
    err = clFinish(commands);
    if (err != CL_SUCCESS) {
      printf("Error: Failed to finish kernel! %s\n", print_cl_errstring(err));
      exit(1);
    }
  }
  *time_kernel+=toc(&tp);
  // Read back the results from the device to verify the output
  tic(&tp);
  err = clEnqueueReadBuffer( commands, urcomp_gfxPtr, CL_TRUE, 0, sizeof(double) * n,
                             (void *)ur, 0, NULL, NULL );
  if (err != CL_SUCCESS) {
    printf("Error: Failed to read real output array! %s\n", print_cl_errstring(err));
    exit(1);
  }
  err = clEnqueueReadBuffer( commands, uicomp_gfxPtr, CL_TRUE, 0, sizeof(double) * n,
                             (void *)ui, 0, NULL, NULL );
  if (err != CL_SUCCESS) {
    printf("Error: Failed to read imaginary output array! %s\n", print_cl_errstring(err));
    exit(1);
  }
  *time_copy+=toc(&tp);
#ifdef TEST


  double *urcomp_gold = (double *)calloc(n, sizeof(double));
  double *uicomp_gold = (double *)calloc(n, sizeof(double));

  const double pi = 3.141592653589793;

  int i, j, ln, ku;

  double t, ti;

  urcomp_gold[0] = 64*nu + m;

  ku    = 1;
  ln    = 1;

  for( j = 0; j <= (m+1)/2 - 1; j++ ) {
    t = pi/(2*ln);
    for( i = 0; i <= ln - 1; i++ ) {
      ti = i*t;
      urcomp_gold[i+ku] = cos( ti );
      uicomp_gold[i+ku] = sin( ti );
    }
    ku = ku + ln;
    ln = 4*ln;
  }

  int ok=0;

  for (i=0; i<n; i++) {
    double eps_r = abs(urcomp_gold[i] - ur[i]);
    double eps_i = abs(uicomp_gold[i] - ui[i]);

    double eps = FEPS;

    if ((eps_r < eps) && (eps_i < eps)) {

    }
    else {
      printf("[%03d] %g / %g -- %g / %g\n", i, urcomp_gold[i], ur[i], uicomp_gold[i], ui[i]);
      ok++;
    }
  }

  printf("init ok: %d\n", ok);

  free(urcomp_gold);
  free(uicomp_gold);
#endif
  clReleaseKernel(kernel);
  clReleaseMemObject(urcomp_gfxPtr);
  clReleaseMemObject(uicomp_gfxPtr);
  return;
  }
  // ---------------------------------------------------------------------
  // --- Transformation call to CFFT4: set initial parameters.

  n  = (int)pow( 2.0, m );
  k = ur0; //ucomp[0].x;
  mx = k%64;

  n1 = n/4;

#ifdef TEST
  n2 = 2*n1;
  n3 = 3*n1;
#endif
  ln = n1;

  // ---------------------------------------------------------------------
  // --- Check for invalid input parameters.

  if ( ( is != 1 && is != -1 ) || m < 2 || m > mx ) {
    printf( " *** Error in CFFT4: Either U has not been initialised\n" );
    printf( "  or one of the input parameters IS or M is invalid:\n" );
    printf( " is = %5d; m = %5d\n", is, m );
    abort();
  }

  // ---------------------------------------------------------------------
  // Start with OpenCL here ...
  // ---------------------------------------------------------------------

  // Create the compute kernel in the program
  kernel = clCreateKernel(program, "cfft4_opencl_r1_kernel", &err);
  if (!kernel || err != CL_SUCCESS) {
    printf("Error: Failed to create compute kernel!\n");
    exit(1);
  }

  // Create and copy to the device memory vectors
  tic(&tp);
  cl_mem xrcomp_gfxPtr = clCreateBuffer(context, CL_MEM_READ_WRITE|CL_MEM_COPY_HOST_PTR,
                                        sizeof(cl_double) * n, xr, NULL);
  cl_mem xicomp_gfxPtr = clCreateBuffer(context, CL_MEM_READ_WRITE|CL_MEM_COPY_HOST_PTR,
                                        sizeof(cl_double) * n, xi, NULL);
  if (!xrcomp_gfxPtr || !xicomp_gfxPtr) {
    printf("Error: Failed to allocate device memory 2!\n");
    exit(1);
  }

  cl_mem wrcomp_gfxPtr = clCreateBuffer(context, CL_MEM_READ_WRITE|CL_MEM_COPY_HOST_PTR,
                                        sizeof(cl_double) * n, wr, NULL);
  cl_mem wicomp_gfxPtr = clCreateBuffer(context, CL_MEM_READ_WRITE|CL_MEM_COPY_HOST_PTR,
                                        sizeof(cl_double) * n, wi, NULL);
  if (!wrcomp_gfxPtr || !wicomp_gfxPtr) {
    printf("Error: Failed to allocate device memory 3!\n");
    exit(1);
  }
  *time_copy+=toc(&tp);

  tic(&tp);
  // Set the arguments to the compute kernel
  err = 0;
  err  = clSetKernelArg(kernel, 0, sizeof(int), &is);
  err |= clSetKernelArg(kernel, 1, sizeof(unsigned long), &n);
  err |= clSetKernelArg(kernel, 2, sizeof(cl_mem), (void*) &xrcomp_gfxPtr);
  err |= clSetKernelArg(kernel, 3, sizeof(cl_mem), (void*) &xicomp_gfxPtr);
  err |= clSetKernelArg(kernel, 4, sizeof(cl_mem), (void*) &wrcomp_gfxPtr);
  err |= clSetKernelArg(kernel, 5, sizeof(cl_mem), (void*) &wicomp_gfxPtr);
  if (err != CL_SUCCESS) {
    printf("Error: Failed to set kernel arguments! %s\n", print_cl_errstring(err));
    exit(1);
  }

  const int fak = BLOCK_SIZE / 8;

  // n/4 == BASISBLOCK
  // /4  bei  32 threads (32/8)
  // /16 bei 128 threads (128/8)
  int nr_local = BLOCK_SIZE;
  int no_blocks = (n / 4)/fak;
  no_blocks += (n/4)%fak?1:0;

  if (no_blocks > 65535) {
    no_blocks = 65535;
  }

  size_t globalWorkers[] = { nr_local * no_blocks };
  size_t localWorkers[] = { nr_local };
  if (no_blocks != 0) {
    err = clEnqueueNDRangeKernel(commands ,kernel,1,0,globalWorkers,
        localWorkers,0,0, NULL);

    if (err) {
      printf("Error: Failed to execute kernel! %s\n", print_cl_errstring(err));
      exit(1);
    }
    clFinish(commands);
  }
  *time_kernel+=toc(&tp);
#ifdef TEST
  // ---------------------------------------------------------------------
  // --- For the first two radix-4 iterations, the inner loop increments J
  //     instead of K. Non-unit stride memory access is used but the longer
  //     vector lengths should compensate for that.

  // Read back the results from the device to verify the output

  err = clEnqueueReadBuffer( commands, wrcomp_gfxPtr, CL_TRUE, 0, sizeof(double) * n,
                             (void *)wr, 0, NULL, NULL );
  if (err != CL_SUCCESS) {
    printf("Error: Failed to read rl output array! %s\n", print_cl_errstring(err));
    exit(1);
  }

  err = clEnqueueReadBuffer( commands, wicomp_gfxPtr, CL_TRUE, 0, sizeof(double) * n,
                             (void *)wi, 0, NULL, NULL );
  if (err != CL_SUCCESS) {
    printf("Error: Failed to read imaginary output array! %s\n", print_cl_errstring(err));
    exit(1);
  }
  double *wrcomp_gold = (double *) calloc(n, sizeof(double));
  double *wicomp_gold = (double *) calloc(n, sizeof(double));

  j0 = 0;
  j1 = j0 + 1;
  j2 = j0 + 2;
  j3 = j0 + 3;
  k0 = j0;
  k1 = k0 + n1;
  k2 = k0 + n2;
  k3 = k0 + n3;
  for( j = 0; j <= ln - 1; j++ ) {
    c0r = xr[j+k0];
    c0i = xi[j+k0];
    c1r = xr[j+k1];
    c1i = xi[j+k1];
    c2r = xr[j+k2];
    c2i = xi[j+k2];
    c3r = xr[j+k3];
    c3i = xi[j+k3];


    d0r = c0r + c2r;
    d0i = c0i + c2i;
    d1r = c0r - c2r;
    d1i = c0i - c2i;
    d2r = c1r + c3r;
    d2i = c1i + c3i;
    d3r = is*(-c1i + c3i);
    d3i = is*( c1r - c3r);

    j4  = 4*j;

	wrcomp_gold[j4+j0] = d0r + d2r;
	wicomp_gold[j4+j0] = d0i + d2i;
	wrcomp_gold[j4+j1] = d1r + d3r;
	wicomp_gold[j4+j1] = d1i + d3i;
	wrcomp_gold[j4+j2] = d0r - d2r;
	wicomp_gold[j4+j2] = d0i - d2i;
	wrcomp_gold[j4+j3] = d1r - d3r;
	wicomp_gold[j4+j3] = d1i - d3i;
    }

  int ok=0;

  for (i=0; i<n; i++) {
    double eps_r = abs(wrcomp_gold[i] - wr[i]);
    double eps_i = abs(wicomp_gold[i] - wi[i]);

    double eps = FEPS; //DBL_EPSILON;

    if ((eps_r < eps) && (eps_i < eps)) {

    }
    else {
     printf("[%03d] %10.4g / %10.4g -- %10.4g / %10.4g -- %10.4g / %10.4g\n", i, wrcomp_gold[i], wr[i], wicomp_gold[i], wi[i], xr[i], xi[i]);
      ok++;
    }
  }

  printf("r1 ok: %d\n", ok);

  free(wrcomp_gold);
  free(wicomp_gold);
#endif
  clReleaseKernel(kernel);

  // ---------------------------------------------------------------------
  // --- If M = 2 all transformation work is done, just copy WR/WI into
  //     XR/XI (label L200).
  //     If M = 3, one radix-2 iteration still has to be done (label L300).

  //if ( m == 2 ) goto L200;
  //if ( m == 3 ) goto L300;

  // --- Else:
  // ----------------------------------------------------------------------
  // --- Perform the second radix-4 iteration.

  ln = ln/4;

  // Create the compute kernel in the program
  kernel = clCreateKernel(program, "cfft4_opencl_r2_kernel", &err);
  if (!kernel || err != CL_SUCCESS) {
    printf("Error: Failed to create compute kernel!\n");
    exit(1);
  }
  tic(&tp);
  // Create and copy to the device memory vectors
  urcomp_gfxPtr = clCreateBuffer(context, CL_MEM_READ_ONLY|CL_MEM_COPY_HOST_PTR,
      sizeof(cl_double) * n, ur, NULL);
  uicomp_gfxPtr = clCreateBuffer(context, CL_MEM_READ_ONLY|CL_MEM_COPY_HOST_PTR,
      sizeof(cl_double) * n, ui, NULL);
  if (!urcomp_gfxPtr || !uicomp_gfxPtr) {
    printf("Error: Failed to allocate device memory 4!\n");
    exit(1);
  }
  *time_copy+=toc(&tp);

  nr_local = BLOCK_SIZE;
  no_blocks_r2 = ln / fak;
  no_blocks_r2 += ln%fak?1:0;

  if (no_blocks_r2 > 65535) {
    no_blocks_r2 = 65535;
  }

  size_t dim = 3;
  size_t r2globalWorkers[3] = { no_blocks_r2 * nr_local , 4  , 1  };
  size_t r2localWorkers[3]  = { nr_local, 1, 1 };

  tic(&tp);
  err = 0;
  err  = clSetKernelArg(kernel, 0, sizeof(int), &is);
  err |= clSetKernelArg(kernel, 1, sizeof(int), &ln);
  err |= clSetKernelArg(kernel, 2, sizeof(int), &n);
  err |= clSetKernelArg(kernel, 3, sizeof(cl_mem), (void*) &urcomp_gfxPtr);
  err |= clSetKernelArg(kernel, 4, sizeof(cl_mem), (void*) &uicomp_gfxPtr);
  err |= clSetKernelArg(kernel, 5, sizeof(cl_mem), (void*) &xrcomp_gfxPtr);
  err |= clSetKernelArg(kernel, 6, sizeof(cl_mem), (void*) &xicomp_gfxPtr);
  err |= clSetKernelArg(kernel, 7, sizeof(cl_mem), (void*) &wrcomp_gfxPtr);
  err |= clSetKernelArg(kernel, 8, sizeof(cl_mem), (void*) &wicomp_gfxPtr);
  if (err != CL_SUCCESS) {
    printf("Error: Failed to set kernel arguments! %s\n", print_cl_errstring(err));
    exit(1);
  }

  if (no_blocks_r2 != 0)
  {
    err = clEnqueueNDRangeKernel(commands, kernel, dim, 0, r2globalWorkers,
        r2localWorkers, 0, 0, NULL);

    if (err) {
      printf("Error: Failed to execute kernel! %s\n", print_cl_errstring(err));
      exit(1);
    }
    clFinish(commands);
  }
  *time_kernel+=toc(&tp);
#ifdef TEST

  // Read back the results from the device to verify the output

  err = clEnqueueReadBuffer( commands, xrcomp_gfxPtr, CL_TRUE, 0, sizeof(double) * n,
      (void *)xr, 0, NULL, NULL );
  if (err != CL_SUCCESS) {
    printf("Error: Failed to read rl output array! %s\n", print_cl_errstring(err));
    exit(1);
  }

  err = clEnqueueReadBuffer( commands, xicomp_gfxPtr, CL_TRUE, 0, sizeof(double) * n,
      (void *)xi, 0, NULL, NULL );
  if (err != CL_SUCCESS) {
    printf("Error: Failed to read imaginary output array! %s\n", print_cl_errstring(err));
    exit(1);
  }

  double *xrcomp_gold = (double *) calloc(n, sizeof(double));
  double *xicomp_gold = (double *) calloc(n, sizeof(double));

  for( k = 0; k <= 3; k++ )
  {

    j0  = k;
    j1  = j0 + 4;
    j2  = j0 + 8;
    j3  = j0 + 12;
    k0  = j0;
    k1  = k0 + n1;
    k2  = k0 + n2;
    k3  = k0 + n3;

    u1r = ur[k+2]; //   ur[k+2];
    u1i = is*ui[k+2]; //is*ui[k+2];
    u2r = u1r*u1r - u1i*u1i;
    u2i = 2.0*u1r*u1i;
    u3r = u1r*u2r - u1i*u2i;
    u3i = u1r*u2i + u1i*u2r;

    for( j = 0; j <= ln - 1; j++ )
    {

      j4  = 4*j;                    

      c0r = wr[j4+k0];         //wr[j4+k0];              // th0
      c0i = wi[j4+k0];         // th1

      w1  = wr[j4+k1];         //wr[j4+k1];              // th2
      w2  = wi[j4+k1];         //wi[j4+k1];              // th3

      c1r = u1r*w1 - u1i*w2;   // th2
      c1i = u1r*w2 + u1i*w1;   // th3

      w1  = wr[j4+k2];         //wr[j4+k2];              // th4
      w2  = wi[j4+k2];         //wi[j4+k2];              // th5

      c2r = u2r*w1 - u2i*w2;   // th4
      c2i = u2r*w2 + u2i*w1;   // th5

      w1  = wr[j4+k3];         //wr[j4+k3];              // th6
      w2  = wi[j4+k3];         //wi[j4+k3];              // th7
      c3r = u3r*w1 - u3i*w2;   // th6
      c3i = u3r*w2 + u3i*w1;   // th7

      // klar
      d0r = c0r + c2r;
      d0i = c0i + c2i;
      d1r = c0r - c2r;
      d1i = c0i - c2i;
      d2r = c1r + c3r;
      d2i = c1i + c3i;
      d3r = is*(-c1i + c3i);
      d3i = is*( c1r - c3r);

      j5  = 16*j;

      xrcomp_gold[j5+j0] = d0r + d2r;
      xicomp_gold[j5+j0] = d0i + d2i;
      xrcomp_gold[j5+j1] = d1r + d3r;
      xicomp_gold[j5+j1] = d1i + d3i;
      xrcomp_gold[j5+j2] = d0r - d2r;
      xicomp_gold[j5+j2] = d0i - d2i;
      xrcomp_gold[j5+j3] = d1r - d3r;
      xicomp_gold[j5+j3] = d1i - d3i;

    }
  }

  ok2=0;

  for (i=0; i<n; i++) {
    double eps_r = abs(xrcomp_gold[i] - xr[i]);
    double eps_i = abs(xicomp_gold[i] - xi[i]);

    double eps = FEPS; //15*DBL_EPSILON;

    if ((eps_r < eps) && (eps_i < eps)) {

    }
    else {
      printf("[%03d] %10.4g / %10.4g -- %10.4g / %10.4g\n", i, xrcomp_gold[i], xr[i], xicomp_gold[i], xi[i]);
      ok2++;
    }
  }

  printf("r2 ok: %d\n", ok2);

  free(xrcomp_gold);
  free(xicomp_gold);
#endif
  clReleaseKernel(kernel);
  ku = 7;
  ln = ln/4;

  l1 = 16;
  l2 = 32;
#ifdef TEST
  l3 = 48;
#endif
  l4 = 64;
  // ---------------------------------------------------------------------
  // --- For all other radix-4 iterations, the inner loops increment K. In
  //     this manner, all array references have unit stride. The radix-4
  //     iterations are performed in pairs: XR/XI to WR/WI and vice
  //     versa. This eliminates the need to copy data between X and W
  //     after each iteration.
  // --- If IT = M - 1 then only one radix-2 iteration needs to be done
  //     (label L100).


  for( it = 4; it <= m - 1; it+= 4 )
  {
    if ( it == m - 1 ) {

      goto L100;
    }
  // Create the compute kernel in the program
  kernel = clCreateKernel(program, "cfft4_opencl_r3_kernel", &err);
  if (!kernel || err != CL_SUCCESS) {
    printf("Error: Failed to create compute kernel (r3)!\n");
    exit(1);
  }

    int x = 0, y = ln , z = 1;

    if (ln > 65535) {
      y=65535;
    }

    int no_blocks_r3 = l1 / fak;
    no_blocks_r3 += l1%fak?1:0;

    if (no_blocks_r3 > 65535) {
      no_blocks_r3 = 65535;
    }

    x = no_blocks_r3;

    size_t r3globalWorkers[3] = { no_blocks_r3 * nr_local , y  , z  };
    size_t r3localWorkers[3]  = { nr_local, 1, 1 };
    tic(&tp);
    err = 0;
    err  = clSetKernelArg(kernel, 0, sizeof(int), &is);
    err |= clSetKernelArg(kernel, 1, sizeof(int), &n);
    err |= clSetKernelArg(kernel, 2, sizeof(int), &ln);
    err |= clSetKernelArg(kernel, 3, sizeof(int), &ku);
    err |= clSetKernelArg(kernel, 4, sizeof(int), &l1);
    err |= clSetKernelArg(kernel, 5, sizeof(cl_mem), (void*) &urcomp_gfxPtr);
    err |= clSetKernelArg(kernel, 6, sizeof(cl_mem), (void*) &uicomp_gfxPtr);
    err |= clSetKernelArg(kernel, 7, sizeof(cl_mem), (void*) &xrcomp_gfxPtr);
    err |= clSetKernelArg(kernel, 8, sizeof(cl_mem), (void*) &xicomp_gfxPtr);
    err |= clSetKernelArg(kernel, 9, sizeof(cl_mem), (void*) &wrcomp_gfxPtr);
    err |= clSetKernelArg(kernel, 10, sizeof(cl_mem), (void*) &wicomp_gfxPtr);

    if (err != CL_SUCCESS) {
      printf("Error: Failed to set kernel arguments! %s\n", print_cl_errstring(err));
      exit(1);
    }
    if (no_blocks_r3 != 0)
    {
      err = clEnqueueNDRangeKernel(commands, kernel, dim, 0, r3globalWorkers,
          r3localWorkers, 0, 0, NULL);

      if (err) {
        printf("Error: Failed to execute kernel! %s\n", print_cl_errstring(err));
        exit(1);
      }
      clFinish(commands);
    }
    *time_kernel+=toc(&tp);
#ifdef TEST

    wrcomp_gold = (double *) calloc(n, sizeof(double));
    wicomp_gold = (double *) calloc(n, sizeof(double));

    // Read back the results from the device to verify the output
    err = clEnqueueReadBuffer( commands, xrcomp_gfxPtr, CL_TRUE, 0, sizeof(double) * n,
        (void *)xr, 0, NULL, NULL );
    if (err != CL_SUCCESS) {
      printf("Error: Failed to read rl output array! %s\n", print_cl_errstring(err));
      exit(1);
    }

    err = clEnqueueReadBuffer( commands, xicomp_gfxPtr, CL_TRUE, 0, sizeof(double) * n,
        (void *)xi, 0, NULL, NULL );
    if (err != CL_SUCCESS) {
      printf("Error: Failed to read imaginary output array! %s\n", print_cl_errstring(err));
      exit(1);
    }
	
    // Read back the results from the device to verify the output

    err = clEnqueueReadBuffer( commands, urcomp_gfxPtr, CL_TRUE, 0, sizeof(double) * n,
        (void *)ur, 0, NULL, NULL );
    if (err != CL_SUCCESS) {
      printf("Error: Failed to read rl output array! %s\n", print_cl_errstring(err));
      exit(1);
    }

    err = clEnqueueReadBuffer( commands, uicomp_gfxPtr, CL_TRUE, 0, sizeof(double) * n,
        (void *)ui, 0, NULL, NULL );
    if (err != CL_SUCCESS) {
      printf("Error: Failed to read imaginary output array! %s\n", print_cl_errstring(err));
      exit(1);
    }
    // Read back the results from the device to verify the output

    err = clEnqueueReadBuffer( commands, wrcomp_gfxPtr, CL_TRUE, 0, sizeof(double) * n,
        (void *)wr, 0, NULL, NULL );
    if (err != CL_SUCCESS) {
      printf("Error: Failed to read rl output array! %s\n", print_cl_errstring(err));
      exit(1);
    }

    err = clEnqueueReadBuffer( commands, wicomp_gfxPtr, CL_TRUE, 0, sizeof(double) * n,
        (void *)wi, 0, NULL, NULL );
    if (err != CL_SUCCESS) {
      printf("Error: Failed to read imaginary output array! %s\n", print_cl_errstring(err));
      exit(1);
    }
	
	for(  j = 0; j <= ln - 1; j++ ) {

	    j0 = j*l4;
	    j1 = j0 + l1;
	    j2 = j0 + l2;
	    j3 = j0 + l3;
	      
	    k0 = j*l1;
	    k1 = k0 + n1;
	    k2 = k0 + n2;
	    k3 = k0 + n3;

	    for( k = 0; k <= l1 - 1; k++ ) {

		u1r =  ur[k+ku-1];   //ur[k+ku-1];
		u1i = is*ui[k+ku-1]; //is*ui[k+ku-1];
		u2r = u1r*u1r - u1i*u1i;
		u2i = 2.0*u1r*u1i;
		u3r = u1r*u2r - u1i*u2i;
		u3i = u1r*u2i + u1i*u2r;

		c0r = xr[k+k0];
		c0i = xi[k+k0];

        x1  = xr[k+k1];   //xr[k+k1];
		x2  = xi[k+k1];   //xi[k+k1];		  

        c1r = u1r*x1 - u1i*x2;      
		c1i = u1r*x2 + u1i*x1;		  
		x1  = xr[k+k2];   //xr[k+k2];
		x2  = xi[k+k2];   //xi[k+k2];	

		c2r = u2r*x1 - u2i*x2;     
		c2i = u2r*x2 + u2i*x1;     	  
		x1  = xr[k+k3];  //xr[k+k3];
		x2  = xi[k+k3];  //xi[k+k3];

		c3r = u3r*x1 - u3i*x2;   
		c3i = u3r*x2 + u3i*x1;      

		d0r = c0r + c2r;         // 0  4
		d0i = c0i + c2i;         // 1  5
		d1r = c0r - c2r;         // 0  4
		d1i = c0i - c2i;         // 1  5
		d2r = c1r + c3r;         // 2  6
		d2i = c1i + c3i;         // 3  7
		d3r = is*(-c1i + c3i);   // 3  7
		d3i = is*( c1r - c3r);   // 2  6	       

		wrcomp_gold[k+j0] = d0r + d2r;    // 0  4
		wicomp_gold[k+j0] = d0i + d2i;    // 1  5
		wrcomp_gold[k+j1] = d1r + d3r;    // 2  6
		wicomp_gold[k+j1] = d1i + d3i;    // 3  7
		wrcomp_gold[k+j2] = d0r - d2r;    // 0  4
		wicomp_gold[k+j2] = d0i - d2i;    // 1  5 
		wrcomp_gold[k+j3] = d1r - d3r;    // 2  6
		wicomp_gold[k+j3] = d1i - d3i;    // 3  7

	    }
	}

	int ok=0;

    for (i=0; i<n; i++)
    {
      double eps_r = abs(wrcomp_gold[i] - wr[i]);
      double eps_i = abs(wicomp_gold[i] - wi[i]);

      double eps = FEPS; //18*DBL_EPSILON;

      if ((eps_r < eps) && (eps_i < eps)) {

      }
      else {
        printf("[%03d] %10.4g / %10.4g -- %10.4g / %10.4g\n", i, wrcomp_gold[i], wr[i], wicomp_gold[i], wi[i]);
        ok++;
      }
    }

	printf("r3 ok: %d\n", ok);
	free(wrcomp_gold);
    free(wicomp_gold);
#endif
  clReleaseKernel(kernel);
	ku = ku + l1;
	ln = ln/4;


	l1 = l4; 
	l2 = 2*l1;
#ifdef TEST
	l3 = 3*l1;
#endif
	l4 = 4*l1;

	//  ----------------------------------------------------------------------
	// --- If IT = M - 2 the computation is complete except for copying
	//                   WR/WI to XR/XI (label L200).
	// --- If IT = M - 3 then only one radix-2 iteration remains to be done
	//                   (label L300).

	if ( it == m - 2 ) {
	    // copy data from gfx-card

	    goto L200;
	}
	if ( it == m - 3 ) {
	    // copy data from gfx-card

	    goto L300;
	}
    // Create the compute kernel in the program
    kernel = clCreateKernel(program, "cfft4_opencl_r4_kernel", &err);
    if (!kernel || err != CL_SUCCESS) {
      printf("Error: Failed to create compute kernel (r4)!\n");
      exit(1);
    }

    x = 0, y = ln , z = 1;

	x=0; y=ln;
	if (ln > 65535) {
	    y=65535;
	}

	int no_blocks_r4 = l1/fak;
	no_blocks_r4 += l1%fak?1:0;

    if (no_blocks_r4 > 65535) {
      no_blocks_r4 = 65535;
    }

    x=no_blocks_r4;

    size_t r4globalWorkers[3] = { no_blocks_r4 * nr_local , y  , z  };
    size_t r4localWorkers[3]  = { nr_local, 1, 1 };
    tic(&tp);
    err = 0;
    err  = clSetKernelArg(kernel, 0, sizeof(int), &is);
    err |= clSetKernelArg(kernel, 1, sizeof(int), &n);
    err |= clSetKernelArg(kernel, 2, sizeof(int), &ln);
    err |= clSetKernelArg(kernel, 3, sizeof(int), &ku);
    err |= clSetKernelArg(kernel, 4, sizeof(int), &l1);
    err |= clSetKernelArg(kernel, 5, sizeof(cl_mem), (void*) &urcomp_gfxPtr);
    err |= clSetKernelArg(kernel, 6, sizeof(cl_mem), (void*) &uicomp_gfxPtr);
    err |= clSetKernelArg(kernel, 7, sizeof(cl_mem), (void*) &xrcomp_gfxPtr);
    err |= clSetKernelArg(kernel, 8, sizeof(cl_mem), (void*) &xicomp_gfxPtr);
    err |= clSetKernelArg(kernel, 9, sizeof(cl_mem), (void*) &wrcomp_gfxPtr);
    err |= clSetKernelArg(kernel, 10, sizeof(cl_mem), (void*) &wicomp_gfxPtr);

    if (err != CL_SUCCESS) {
      printf("Error: Failed to set kernel arguments (r4)! %s\n", print_cl_errstring(err));
      exit(1);
    }
    if (no_blocks_r4 != 0)
    {
      err = clEnqueueNDRangeKernel(commands, kernel, dim, 0, r4globalWorkers,
          r4localWorkers, 0, 0, NULL);

      if (err) {
        printf("Error: Failed to execute kernel! %s\n", print_cl_errstring(err));
        exit(1);
      }
      clFinish(commands);
    }
    *time_kernel+=toc(&tp);
    // Read back the results from the device to verify the output
    tic(&tp);
    err = clEnqueueReadBuffer( commands, xrcomp_gfxPtr, CL_TRUE, 0, sizeof(double) * n,
        (void *)xr, 0, NULL, NULL );
    if (err != CL_SUCCESS) {
      printf("Error: Failed to read reall output array (r4)! %s\n", print_cl_errstring(err));
      exit(1);
    }

    err = clEnqueueReadBuffer( commands, xicomp_gfxPtr, CL_TRUE, 0, sizeof(double) * n,
        (void *)xi, 0, NULL, NULL );
    if (err != CL_SUCCESS) {
      printf("Error: Failed to read imaginary output array! %s\n", print_cl_errstring(err));
      exit(1);
    }
    *time_copy+=toc(&tp);
#ifdef TEST

    xrcomp_gold = (double *) calloc(n, sizeof(double));
    xicomp_gold = (double *) calloc(n, sizeof(double));

	for( j = 0; j <= ln - 1; j++ )
    {
      j0 = j*l4;
      j1 = j0 + l1;
      j2 = j0 + l2;
      j3 = j0 + l3;
      k0 = j*l1;
      k1 = k0 + n1;
      k2 = k0 + n2;
      k3 = k0 + n3;

      for( k = 0; k <= l1 - 1; k++ ) {

        u1r = ur[k+ku-1];     // ur[k+ku-1];
        u1i = is*ui[k+ku-1];  //is*ui[k+ku-1];
        u2r = u1r*u1r - u1i*u1i;
        u2i = 2.0*u1r*u1i;
        u3r = u1r*u2r - u1i*u2i;
        u3i = u1r*u2i + u1i*u2r;

        c0r = wr[k+k0];
        c0i = wi[k+k0]; //wi[k+k0];

        w1  = wr[k+k1]; //wr[k+k1];
        w2  = wi[k+k1]; //wi[k+k1];

        c1r = u1r*w1 - u1i*w2;
        c1i = u1r*w2 + u1i*w1;
        w1  = wr[k+k2]; //wr[k+k2];
        w2  = wi[k+k2]; //wi[k+k2];

        c2r = u2r*w1 - u2i*w2;
        c2i = u2r*w2 + u2i*w1;

        w1  = wr[k+k3]; //wr[k+k3];
        w2  = wi[k+k3]; //wi[k+k3];

        c3r = u3r*w1 - u3i*w2;
        c3i = u3r*w2 + u3i*w1;

        d0r = c0r + c2r;
        d0i = c0i + c2i;
        d1r = c0r - c2r;
        d1i = c0i - c2i;
        d2r = c1r + c3r;
        d2i = c1i + c3i;
        d3r = is*(-c1i + c3i);
        d3i = is*( c1r - c3r);

        xrcomp_gold[k+j0] = d0r + d2r;    // 0  4
        xicomp_gold[k+j0] = d0i + d2i;    // 1  5
        xrcomp_gold[k+j1] = d1r + d3r;    // 2  6
        xicomp_gold[k+j1] = d1i + d3i;    // 3  7
        xrcomp_gold[k+j2] = d0r - d2r;    // 0  4
        xicomp_gold[k+j2] = d0i - d2i;    // 1  5 
        xrcomp_gold[k+j3] = d1r - d3r;    // 2  6
        xicomp_gold[k+j3] = d1i - d3i;    // 3  7
      }
    }

	int ok4=0;

    for (i=0; i<n; i++)
    {
      double eps_r = abs(xrcomp_gold[i] - xr[i]);
      double eps_i = abs(xicomp_gold[i] - xi[i]);

      double eps = FEPS; //19*DBL_EPSILON;

      if ((eps_r < eps) && (eps_i < eps)) {
      }
      else {
        printf("[%03d] %10.4g / %10.4g -- %10.4g / %10.4g\n", i, xrcomp_gold[i], xr[i], xicomp_gold[i], xi[i]);
        ok4++;
      }
    }

    printf("r4 ok: %d\n", ok4);

    free(xrcomp_gold);
    free(xicomp_gold);

#endif

    ku = ku + l1;
	ln = ln/4;
	l1 = l4;
	l2 = 2*l1;
#ifdef TEST
	l3 = 3*l1;
#endif
	l4 = 4*l1;
  clReleaseKernel(kernel);
  }

  clReleaseMemObject(xrcomp_gfxPtr);
  clReleaseMemObject(xicomp_gfxPtr);
  clReleaseMemObject(urcomp_gfxPtr);
  clReleaseMemObject(uicomp_gfxPtr);
  clReleaseMemObject(wrcomp_gfxPtr);
  clReleaseMemObject(wicomp_gfxPtr);
  return;
L100:

  l2 = l1/2;

  // Create the compute kernel in the program
  kernel = clCreateKernel(program, "cfft4_opencl_r5_kernel", &err);
  if (!kernel || err != CL_SUCCESS) {
    printf("Error: Failed to create compute kernel (r5)! %s\n", print_cl_errstring(err));
    exit(1);
  }

  no_blocks_r5 = l2/fak;
  no_blocks_r5 += l2%fak?1:0;

  if (no_blocks_r5 > 65535) {
    no_blocks_r5 = 65535;
  }
  size_t r5globalWorkers[] = { no_blocks_r5 * nr_local };
  size_t r5localWorkers[]  = { nr_local };
  tic(&tp);
  err = 0;
  err  = clSetKernelArg(kernel, 0, sizeof(int), &is);
  err |= clSetKernelArg(kernel, 1, sizeof(int), &l1);
  err |= clSetKernelArg(kernel, 2, sizeof(int), &ku);
  err |= clSetKernelArg(kernel, 3, sizeof(cl_mem), (void*) &urcomp_gfxPtr);
  err |= clSetKernelArg(kernel, 4, sizeof(cl_mem), (void*) &uicomp_gfxPtr);
  err |= clSetKernelArg(kernel, 5, sizeof(cl_mem), (void*) &xrcomp_gfxPtr);
  err |= clSetKernelArg(kernel, 6, sizeof(cl_mem), (void*) &xicomp_gfxPtr);
  err |= clSetKernelArg(kernel, 7, sizeof(cl_mem), (void*) &wrcomp_gfxPtr);
  err |= clSetKernelArg(kernel, 8, sizeof(cl_mem), (void*) &wicomp_gfxPtr);

  if (err != CL_SUCCESS) {
    printf("Error: Failed to set kernel arguments (r5)! %s\n", print_cl_errstring(err));
    exit(1);
  }
  if (no_blocks_r5 != 0)
  {
    err = clEnqueueNDRangeKernel(commands, kernel, 1, 0, r5globalWorkers,
        r5localWorkers, 0, 0, NULL);

    if (err) {
      printf("Error: Failed to execute kernel! %s\n", print_cl_errstring(err));
      exit(1);
    }
    clFinish(commands);
  }
  *time_kernel+=toc(&tp);
#ifdef TEST

    wrcomp_gold = (double *) calloc(n, sizeof(double));
    wicomp_gold = (double *) calloc(n, sizeof(double));
    // Read back the results from the device to verify the output

    err = clEnqueueReadBuffer( commands, wrcomp_gfxPtr, CL_TRUE, 0, sizeof(double) * n,
        (void *)wr, 0, NULL, NULL );
    if (err != CL_SUCCESS) {
      printf("Error: Failed to read rl output array! %s\n", print_cl_errstring(err));
      exit(1);
    }

    err = clEnqueueReadBuffer( commands, wicomp_gfxPtr, CL_TRUE, 0, sizeof(double) * n,
        (void *)wi, 0, NULL, NULL );
    if (err != CL_SUCCESS) {
      printf("Error: Failed to read imaginary output array! %s\n", print_cl_errstring(err));
      exit(1);
    }

    l2 = l1/2;
    j0 = 0;
    j1 = j0 + l1;
    j2 = j0 + l2;
    j3 = j1 + l2;
    k0 = 0;
    k1 = k0 + l1;
    k2 = k0 + l2;
    k3 = k1 + l2;
    for( k = 0; k <= l2 - 1; k++ ) {
      u1r =    ur[2*k+ku-1];
      u1i = is*ui[2*k+ku-1];
      c0r = xr[k+k0];
      c0i = xi[k+k0];
      x1  = xr[k+k1];
      x2  = xi[k+k1];
      c1r = u1r*x1 - u1i*x2;
      c1i = u1r*x2 + u1i*x1;
      wrcomp_gold[k+j0] = c0r + c1r;
      wicomp_gold[k+j0] = c0i + c1i;
      wrcomp_gold[k+j1] = c0r - c1r;
      wicomp_gold[k+j1] = c0i - c1i;
      u2r = -is*u1i;
      u2i =  is*u1r;
      c0r = xr[k+k2];
      c0i = xi[k+k2];
      x1  = xr[k+k3];
      x2  = xi[k+k3];
      c1r = u2r*x1 - u2i*x2;
      c1i = u2r*x2 + u2i*x1;
      wrcomp_gold[k+j2] = c0r + c1r;
      wicomp_gold[k+j2] = c0i + c1i;
      wrcomp_gold[k+j3] = c0r - c1r;
      wicomp_gold[k+j3] = c0i - c1i;
    }

    int ok5=0;

    for (i=0; i<n; i++)
    {
      double eps_r = abs(wrcomp_gold[i] - wr[i]);
      double eps_i = abs(wicomp_gold[i] - wi[i]);

      double eps = FEPS; //19*DBL_EPSILON;

      if ((eps_r < eps) && (eps_i < eps)) {
      }
      else {
        printf("[%03d] %10.5g / %10.5g -- %10.5g / %10.5g\n", i, wrcomp_gold[i], wr[i], wicomp_gold[i], wi[i]);
        ok5++;
      }
    }

	printf("r5 ok: %d\n", ok5);
	free(wrcomp_gold);
	free(wicomp_gold);
#endif

L200:
    tic(&tp);
    err = clEnqueueReadBuffer( commands, wrcomp_gfxPtr, CL_TRUE, 0, sizeof(double) * n,
        (void *)xr, 0, NULL, NULL );
    if (err != CL_SUCCESS) {
      printf("Error: Failed to read rl output array! %s\n", print_cl_errstring(err));
      exit(1);
    }

    err = clEnqueueReadBuffer( commands, wicomp_gfxPtr, CL_TRUE, 0, sizeof(double) * n,
        (void *)xi, 0, NULL, NULL );
    if (err != CL_SUCCESS) {
      printf("Error: Failed to read imaginary output array! %s\n", print_cl_errstring(err));
      exit(1);
    }
    *time_copy+=toc(&tp);
    clReleaseMemObject(xrcomp_gfxPtr);
    clReleaseMemObject(xicomp_gfxPtr);
    clReleaseMemObject(urcomp_gfxPtr);
    clReleaseMemObject(uicomp_gfxPtr);
    clReleaseMemObject(wrcomp_gfxPtr);
    clReleaseMemObject(wicomp_gfxPtr);
    return;

L300: ;


  // Create the compute kernel in the program
  kernel = clCreateKernel(program, "cfft4_opencl_r6_kernel", &err);
  if (!kernel || err != CL_SUCCESS) {
    printf("Error: Failed to create compute kernel (r6)! %s\n", print_cl_errstring(err));
    exit(1);
  }

  if ( m == 3 ) {
    l1 = 4;
    ku = 3;
  }

  l2 = l1/2;

  int no_blocks_r6 = l2/fak;
  no_blocks_r6 += l2%fak?1:0;

  if (no_blocks_r6 > 65535)
  {
    no_blocks_r6 = 65535;
  }

  size_t r6globalWorkers[] = { no_blocks_r6 * nr_local };
  size_t r6localWorkers[]  = { nr_local };
  tic(&tp);
  err = 0;
  err  = clSetKernelArg(kernel, 0, sizeof(int), &is);
  err |= clSetKernelArg(kernel, 1, sizeof(int), &l1);
  err |= clSetKernelArg(kernel, 2, sizeof(int), &ku);
  err |= clSetKernelArg(kernel, 3, sizeof(cl_mem), (void*) &urcomp_gfxPtr);
  err |= clSetKernelArg(kernel, 4, sizeof(cl_mem), (void*) &uicomp_gfxPtr);
  err |= clSetKernelArg(kernel, 5, sizeof(cl_mem), (void*) &xrcomp_gfxPtr);
  err |= clSetKernelArg(kernel, 6, sizeof(cl_mem), (void*) &xicomp_gfxPtr);
  err |= clSetKernelArg(kernel, 7, sizeof(cl_mem), (void*) &wrcomp_gfxPtr);
  err |= clSetKernelArg(kernel, 8, sizeof(cl_mem), (void*) &wicomp_gfxPtr);

  if (err != CL_SUCCESS) {
    printf("Error: Failed to set kernel arguments (r6)! %s\n", print_cl_errstring(err));
    exit(1);
  }
  if (no_blocks_r6 != 0)
  {
    err = clEnqueueNDRangeKernel(commands, kernel, 1, 0, r6globalWorkers,
        r6localWorkers, 0, 0, NULL);

    if (err) {
      printf("Error: Failed to execute kernel (r6)! %s\n", print_cl_errstring(err));
      exit(1);
    }
    clFinish(commands);
  }
  *time_kernel+=toc(&tp);
  tic(&tp);
  err = clEnqueueReadBuffer( commands, xrcomp_gfxPtr, CL_TRUE, 0, sizeof(double) * n,
      (void *)xr, 0, NULL, NULL );
  if (err != CL_SUCCESS) {
    printf("Error: Failed to read rl output array! %s\n", print_cl_errstring(err));
    exit(1);
  }

  err = clEnqueueReadBuffer( commands, xicomp_gfxPtr, CL_TRUE, 0, sizeof(double) * n,
      (void *)xi, 0, NULL, NULL );
  if (err != CL_SUCCESS) {
    printf("Error: Failed to read imaginary output array! %s\n", print_cl_errstring(err));
    exit(1);
  }
  *time_copy+=toc(&tp);

#ifdef TEST
  xrcomp_gold = (double *) calloc(n, sizeof(double));
  xicomp_gold = (double *) calloc(n, sizeof(double));
  j0 = 0;
  j1 = j0 + l1;
  j2 = j0 + l2;
  j3 = j1 + l2;
  k0 = 0;
  k1 = k0 + l1;
  k2 = k0 + l2;
  k3 = k1 + l2;

  for( k = 0; k <= l2 - 1; k++ )
  {
    u1r =    ur[2*k+ku-1];
    u1i = is*ui[2*k+ku-1];
    c0r = wr[k+k0];
    c0i = wi[k+k0];
    w1  = wr[k+k1];
    w2  = wi[k+k1];
    c1r = u1r*w1 - u1i*w2;
    c1i = u1r*w2 + u1i*w1;
    xrcomp_gold[k+j0] = c0r + c1r;
    xicomp_gold[k+j0] = c0i + c1i;
    xrcomp_gold[k+j1] = c0r - c1r;
    xicomp_gold[k+j1] = c0i - c1i;
    u2r = -is*u1i;
    u2i =  is*u1r;
    c0r = wr[k+k2];
    c0i = wi[k+k2];
    w1  = wr[k+k3];
    w2  = wi[k+k3];
    c1r = u2r*w1 - u2i*w2;
    c1i = u2r*w2 + u2i*w1;
    xrcomp_gold[k+j2] = c0r + c1r;
    xicomp_gold[k+j2] = c0i + c1i;
    xrcomp_gold[k+j3] = c0r - c1r;
    xicomp_gold[k+j3] = c0i - c1i;
  }

  int ok6=0;

  for (i=0; i<n; i++)
  {
    double eps_r = abs(xrcomp_gold[i] - xr[i]);
    double eps_i = abs(xicomp_gold[i] - xi[i]);

    double eps = FEPS; //19*DBL_EPSILON;

    if ((eps_r < eps) && (eps_i < eps))
    {
    }
    else
    {
      printf("[%03d] %10.5g / %10.5g -- %10.5g / %10.5g\n", i, xrcomp_gold[i], xr[i], xicomp_gold[i], xi[i]);
      ok6++;
    }
  }

  printf("r6 ok: %d\n", ok6);
  free(xrcomp_gold);
  free(xicomp_gold);
#endif
  clReleaseMemObject(xrcomp_gfxPtr);
  clReleaseMemObject(xicomp_gfxPtr);
  clReleaseMemObject(urcomp_gfxPtr);
  clReleaseMemObject(uicomp_gfxPtr);
  clReleaseMemObject(wrcomp_gfxPtr);
  clReleaseMemObject(wicomp_gfxPtr);
return;
}

char *print_cl_errstring(cl_int err)
{
  switch (err)
  {
    case CL_SUCCESS:                          return strdup("Success!");
    case CL_DEVICE_NOT_FOUND:                 return strdup("Device not found.");
    case CL_DEVICE_NOT_AVAILABLE:             return strdup("Device not available");
    case CL_COMPILER_NOT_AVAILABLE:           return strdup("Compiler not available");
    case CL_MEM_OBJECT_ALLOCATION_FAILURE:    return strdup("Memory object allocation failure");
    case CL_OUT_OF_RESOURCES:                 return strdup("Out of resources");
    case CL_OUT_OF_HOST_MEMORY:               return strdup("Out of host memory");
    case CL_PROFILING_INFO_NOT_AVAILABLE:     return strdup("Profiling information not available");
    case CL_MEM_COPY_OVERLAP:                 return strdup("Memory copy overlap");
    case CL_IMAGE_FORMAT_MISMATCH:            return strdup("Image format mismatch");
    case CL_IMAGE_FORMAT_NOT_SUPPORTED:       return strdup("Image format not supported");
    case CL_BUILD_PROGRAM_FAILURE:            return strdup("Program build failure");
    case CL_MAP_FAILURE:                      return strdup("Map failure");
    case CL_INVALID_VALUE:                    return strdup("Invalid value");
    case CL_INVALID_DEVICE_TYPE:              return strdup("Invalid device type");
    case CL_INVALID_PLATFORM:                 return strdup("Invalid platform");
    case CL_INVALID_DEVICE:                   return strdup("Invalid device");
    case CL_INVALID_CONTEXT:                  return strdup("Invalid context");
    case CL_INVALID_QUEUE_PROPERTIES:         return strdup("Invalid queue properties");
    case CL_INVALID_COMMAND_QUEUE:            return strdup("Invalid command queue");
    case CL_INVALID_HOST_PTR:                 return strdup("Invalid host pointer");
    case CL_INVALID_MEM_OBJECT:               return strdup("Invalid memory object");
    case CL_INVALID_IMAGE_FORMAT_DESCRIPTOR:  return strdup("Invalid image format descriptor");
    case CL_INVALID_IMAGE_SIZE:               return strdup("Invalid image size");
    case CL_INVALID_SAMPLER:                  return strdup("Invalid sampler");
    case CL_INVALID_BINARY:                   return strdup("Invalid binary");
    case CL_INVALID_BUILD_OPTIONS:            return strdup("Invalid build options");
    case CL_INVALID_PROGRAM:                  return strdup("Invalid program");
    case CL_INVALID_PROGRAM_EXECUTABLE:       return strdup("Invalid program executable");
    case CL_INVALID_KERNEL_NAME:              return strdup("Invalid kernel name");
    case CL_INVALID_KERNEL_DEFINITION:        return strdup("Invalid kernel definition");
    case CL_INVALID_KERNEL:                   return strdup("Invalid kernel");
    case CL_INVALID_ARG_INDEX:                return strdup("Invalid argument index");
    case CL_INVALID_ARG_VALUE:                return strdup("Invalid argument value");
    case CL_INVALID_ARG_SIZE:                 return strdup("Invalid argument size");
    case CL_INVALID_KERNEL_ARGS:              return strdup("Invalid kernel arguments");
    case CL_INVALID_WORK_DIMENSION:           return strdup("Invalid work dimension");
    case CL_INVALID_WORK_GROUP_SIZE:          return strdup("Invalid work group size");
    case CL_INVALID_WORK_ITEM_SIZE:           return strdup("Invalid work item size");
    case CL_INVALID_GLOBAL_OFFSET:            return strdup("Invalid global offset");
    case CL_INVALID_EVENT_WAIT_LIST:          return strdup("Invalid event wait list");
    case CL_INVALID_EVENT:                    return strdup("Invalid event");
    case CL_INVALID_OPERATION:                return strdup("Invalid operation");
    case CL_INVALID_GL_OBJECT:                return strdup("Invalid OpenGL object");
    case CL_INVALID_BUFFER_SIZE:              return strdup("Invalid buffer size");
    case CL_INVALID_MIP_LEVEL:                return strdup("Invalid mip-map level");
    default:                                  return strdup("Unknown");
  }
} 

