#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>

#include "cfft4_opencl.h"

double cclock( void );
int    check( double a[], double b[], int m, double err );
void   cfft4( int, int, double ur[], double ui[], double xr[], double xi[],
              double wr[], double wi[] );
void   gendat( double *a, int );
void   nflops( unsigned long long, unsigned long long *, unsigned long long * );
void   prthead( void );
void   prtspeed( int, double, double, double, double, double, double, int );
void   state( char * );


int main()
{
  unsigned long  m, n, nrep, ok;
  unsigned long long mflint, mfltrn;
  int      i, j;

  double   *ur, *ui, *wr, *wi;
  double *a, *b, *c;

  double   corr, err;
  double mflops, time;
  double mflops_kernel, mflops_copy_kernel;
  double   time_kernel = 0.0, time_copy = 0.0, time_copy_kernel;
  FILE     *inl;
  // ------------------------------------------------------------------------

  struct timeval tb, te;
  double duration;
  // init OpenCL
  init_opencl(1);
  state( "mod2f" );
  prthead();
  inl = fopen( "mod2f.in", "r" );
  while( ( fscanf( inl, "%lu%lu\n", &m, &nrep ) != EOF ) ){
    time_kernel = 0.0;
    time_copy = 0.0;
    n   = pow( 2, m );
    err = ( 10.0*n*m )* 1.0e-7;         // --- Allowed error tolerance.

    ur  = calloc( n, sizeof( double ) );
    ui  = calloc( n, sizeof( double ) );
    wr  = calloc( n, sizeof( double ) );
    wi  = calloc( n, sizeof( double ) );

    c   = calloc( n, sizeof( double ) );

    a   = calloc( n, sizeof( double ) );
    b   = calloc( n, sizeof( double ) );


    // -------------------------------------------------------------------------
    // --- Generate data.

    gendat( c, n );

    // -------------------------------------------------------------------------
    // --- Repeat FFT 'nrep' times for this problem size.


    time = cclock();
    for( i = 0; i < nrep; i++ ) {
      for( j = 0; j < n; j++ ) {

        a[j] = c[j];
        b[j] = 0.0;

      }
      cfft4_opencl( 0, m, ur, ui, a, b, wr, wi, &time_kernel, &time_copy);
      cfft4_opencl( 1, m, ur, ui, a, b, wr, wi, &time_kernel, &time_copy);
    }
    time = cclock() - time;

    // -------------------------------------------------------------------------
    // --- Check for errors and correct timing for filling of arrays.

    ok   = check( a, b, n, err );
    corr = cclock();

    for( i = 0; i < nrep; i++ ) {
      for( j = 0; j < n; j++ ) {
        a[j] = c[j];
        b[j] = 0.0;
      }
    }

    corr = cclock() - corr;
    time = ( time - corr )/(double)nrep;
    time_kernel/=(double)nrep;
    time_copy/=(double)nrep;
    time_copy_kernel = time_kernel + time_copy;
    // -------------------------------------------------------------------------
    // --- Calculate Mflop rates.

    nflops( m, &mflint, &mfltrn );
    mflops = 1.0e-6*( mflint + mfltrn )/time;
    mflops_kernel = 1.0e-6*( mflint + mfltrn )/time_kernel;
    mflops_copy_kernel = 1.0e-6*( mflint + mfltrn )/time_copy_kernel;

    prtspeed( n, time, mflops, time_copy_kernel, mflops_copy_kernel, time_kernel, mflops_kernel, ok );

    free( wi ); free( wr ); free( ui ); free( ur );

    free( c );

    free( b ); free( a );	

  }
  finish_opencl();
  printf( "-----------------------------------------------------------------------------------------------------------\n" );

  return 0;
}
